theory Formula
imports
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "../Rewriting/Term"
  "../Rewriting/Trs_Impl"
  "../Auxiliaries/Option_Monad"
  "Sorted_Algebra"
begin

type_synonym ('v,'t,'d) valuation = "'v \<times> 't \<Rightarrow> 'd"

datatype 'a formula =
  Atom "'a"
| NegAtom "'a"
| Conjunction "'a formula list"
| Disjunction "'a formula list"

derive "show" formula

fun isLit where "isLit (Atom _) = True" | "isLit (NegAtom _) = True" | "isLit _ = False"

abbreviation disj_atoms :: "'a list \<Rightarrow> 'a formula" where
  "disj_atoms as \<equiv> Disjunction (map Atom as)"
abbreviation conj_atoms :: "'a list \<Rightarrow> 'a formula" where
  "conj_atoms as \<equiv> Conjunction (map Atom as)"

inductive is_conj_lits
where "(\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> isLit \<phi>) \<Longrightarrow> is_conj_lits (Conjunction \<phi>s)"

inductive is_clause
  where "(\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> isLit \<phi>) \<Longrightarrow> is_clause (Disjunction \<phi>s)"
    
lemma is_clause_code[code]: "is_clause (Disjunction ls) = (\<forall> l \<in> set ls. isLit l)" 
  "is_clause (Conjunction xs) = False" 
  "is_clause (Atom a) = False" 
  "is_clause (NegAtom a) = False" 
  by (auto simp: is_clause.simps)

inductive is_cnf
where "(\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> is_clause \<phi>) \<Longrightarrow> is_cnf (Conjunction \<phi>s)"

abbreviation form_False :: "'a formula" ("False\<^sub>f")
where "form_False \<equiv> Disjunction []"

abbreviation form_True  :: "'a formula" ("True\<^sub>f")
where "form_True \<equiv> Conjunction []"

fun form_not :: "'a formula \<Rightarrow> 'a formula" ("\<not>\<^sub>f _" [43] 43)
where "(\<not>\<^sub>f Atom a) = NegAtom a"
    | "(\<not>\<^sub>f NegAtom a) = Atom a"
    | "(\<not>\<^sub>f Conjunction \<phi>s) = Disjunction (map form_not \<phi>s)"
    | "(\<not>\<^sub>f Disjunction \<phi>s) = Conjunction (map form_not \<phi>s)"

fun form_and :: "'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula" (infixl "\<and>\<^sub>f" 42)
where "(Conjunction \<phi>s \<and>\<^sub>f Conjunction \<psi>s) = Conjunction (\<phi>s @ \<psi>s)"
    | "(Conjunction \<phi>s \<and>\<^sub>f \<psi>) = Conjunction (\<phi>s @ [\<psi>])"
    | "(\<phi> \<and>\<^sub>f Conjunction \<psi>s) = Conjunction (\<phi> # \<psi>s)"
    | "(\<phi> \<and>\<^sub>f \<psi>) = Conjunction [\<phi>,\<psi>]"

fun form_or :: "'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula" (infixl "\<or>\<^sub>f" 41)
where "(Disjunction \<phi>s \<or>\<^sub>f Disjunction \<psi>s) = Disjunction (\<phi>s @ \<psi>s)"
    | "(Disjunction \<phi>s \<or>\<^sub>f \<psi>) = Disjunction (\<phi>s @ [\<psi>])"
    | "(\<phi> \<or>\<^sub>f Disjunction \<psi>s) = Disjunction (\<phi> # \<psi>s)"
    | "(\<phi> \<or>\<^sub>f \<psi>) = Disjunction [\<phi>,\<psi>]"

fun cnf_form_or :: "'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula" (infixl "\<or>\<^sub>c\<^sub>f" 41) where 
  "(Conjunction \<phi>s \<or>\<^sub>c\<^sub>f Conjunction \<psi>s) = Conjunction [\<phi> \<or>\<^sub>f \<psi> . \<phi> \<leftarrow> \<phi>s, \<psi> \<leftarrow> \<psi>s]"
| "(\<phi> \<or>\<^sub>c\<^sub>f \<psi>) = (\<phi> \<or>\<^sub>f \<psi>)"


lemma map_form_and[simp]: "map_formula f (\<phi> \<and>\<^sub>f \<psi>) = (map_formula f \<phi> \<and>\<^sub>f map_formula f \<psi>)"
  by (cases \<phi>; cases \<psi>, auto)

lemma map_form_or[simp]: "map_formula f (\<phi> \<or>\<^sub>f \<psi>) = (map_formula f \<phi> \<or>\<^sub>f map_formula f \<psi>)"
  by (cases \<phi>; cases \<psi>, auto)

lemma map_form_not[simp]: "map_formula f (\<not>\<^sub>f \<phi>) = (\<not>\<^sub>f map_formula f \<phi>)"
  by (induct \<phi>, auto)

text {* The internal constructs are less useful. *}

declare form_and.simps[simp del] form_or.simps[simp del]

abbreviation form_imp (infixr "\<longrightarrow>\<^sub>f" 40) where "(\<phi> \<longrightarrow>\<^sub>f \<psi>) \<equiv> \<not>\<^sub>f \<phi> \<or>\<^sub>f \<psi>"

fun form_all
where "form_all [] = True\<^sub>f"
    | "form_all (\<phi>#\<phi>s) = (\<phi> \<and>\<^sub>f form_all \<phi>s)"

fun form_ex
where "form_ex [] = False\<^sub>f"
    | "form_ex (\<phi>#\<phi>s) = (\<phi> \<or>\<^sub>f form_ex \<phi>s)"

fun form_cnf_ex
where "form_cnf_ex [] = Conjunction [False\<^sub>f]"
    | "form_cnf_ex (\<phi>#\<phi>s) = (\<phi> \<or>\<^sub>c\<^sub>f form_cnf_ex \<phi>s)"

fun flatten where
  "flatten (Conjunction \<phi>s) = form_all (map flatten \<phi>s)"
| "flatten (Disjunction \<phi>s) = form_cnf_ex (map flatten \<phi>s)"
| "flatten \<phi> = Conjunction [Disjunction [\<phi>]]"

fun conjuncts where "conjuncts (Conjunction \<phi>s) = \<phi>s" | "conjuncts \<phi> = [\<phi>]"

fun disjuncts where "disjuncts (Disjunction \<phi>s) = \<phi>s" | "disjuncts \<phi> = [\<phi>]"

fun triv_unsat :: "'a formula \<Rightarrow> bool"
where "triv_unsat (Conjunction \<phi>s) = (\<forall>\<phi> \<in> set \<phi>s. triv_unsat \<phi>)"
    | "triv_unsat _ = False"




locale formula =
  fixes atom :: "'atom \<Rightarrow> bool"
    and assignment :: "'assignment \<Rightarrow> bool"
    and eval_true :: "'assignment \<Rightarrow> 'atom \<Rightarrow> bool"
begin

abbreviation "Assignments \<equiv> Collect assignment"

subsection {* Well-formedness *}

fun formula :: "'atom formula \<Rightarrow> bool" where
  "formula (Atom a) = atom a"
| "formula (NegAtom a) = atom a"
| "formula (Conjunction \<phi>s) = (\<forall>\<phi> \<in> set \<phi>s. formula \<phi>)"
| "formula (Disjunction \<phi>s) = (\<forall>\<phi> \<in> set \<phi>s. formula \<phi>)"
 
lemma formula_and[simp]: "formula (\<phi> \<and>\<^sub>f \<psi>) \<longleftrightarrow> formula \<phi> \<and> formula \<psi>"
  by (cases \<phi>; cases \<psi>, auto simp: form_and.simps)

lemma formula_or[simp]: "formula (\<phi> \<or>\<^sub>f \<psi>) \<longleftrightarrow> formula \<phi> \<and> formula \<psi>"
  by (cases \<phi>; cases \<psi>, auto simp: form_or.simps)

lemma formula_not[simp]: "formula (\<not>\<^sub>f \<phi>) \<longleftrightarrow> formula \<phi>" by (induct \<phi>; auto)

lemma formula_all[simp]: "formula (form_all \<phi>s) \<longleftrightarrow> (\<forall>\<phi> \<in> set \<phi>s. formula \<phi>)"
  by (induct \<phi>s, auto)

lemma formula_ex[simp]: "formula (form_ex \<phi>s) \<longleftrightarrow> (\<forall>\<phi> \<in> set \<phi>s. formula \<phi>)"
  by (induct \<phi>s, auto)

lemma formula_cnf_or[simp]: "formula \<phi> \<Longrightarrow> formula \<psi> \<Longrightarrow> formula (\<phi> \<or>\<^sub>c\<^sub>f \<psi>)"
  by (cases \<phi>; cases \<psi>, auto)

lemma formula_cnf_ex[simp]: "(\<forall>\<phi> \<in> set \<phi>s. formula \<phi>) \<Longrightarrow> formula (form_cnf_ex \<phi>s)"
  by (induct \<phi>s, auto)

lemma formula_flatten[simp]: "formula \<phi> \<Longrightarrow> formula (flatten \<phi>)"
  by (induct \<phi> rule: flatten.induct, auto simp: Let_def)

lemma isLit_negate[simp]: "isLit (\<not>\<^sub>f f) = isLit f" by (cases f, auto)

lemma is_conj_lits_negate[simp]: "is_conj_lits (\<not>\<^sub>f f) = is_clause f"
  by (cases f, auto simp: is_clause.simps is_conj_lits.simps)

lemma is_cnf_form_and[simp]: "is_cnf f \<Longrightarrow> is_cnf g \<Longrightarrow> is_cnf (f \<and>\<^sub>f g)"
  by (cases f; cases g; auto simp: is_cnf.simps form_and.simps)

lemma is_clause_form_or[simp]: "is_clause f \<Longrightarrow> is_clause g \<Longrightarrow> is_clause (f \<or>\<^sub>f g)"
  by (cases f; cases g; auto simp: is_clause.simps form_or.simps)

lemma is_cnf_cnf_form_or[simp]: "is_cnf f \<Longrightarrow> is_cnf g \<Longrightarrow> is_cnf (f \<or>\<^sub>c\<^sub>f g)"
  by (cases f; cases g; auto simp: is_cnf.simps form_and.simps)

lemma is_cnf_form_True[simp]: "is_cnf True\<^sub>f"
  by (simp add: is_cnf.simps)

lemma is_cnf_form_all[simp]: "(\<And> f. f \<in> set fs \<Longrightarrow> is_cnf f) \<Longrightarrow> is_cnf (form_all fs)"
  by(induct fs, auto)

lemma is_cnf_form_cnf_ex[simp]: "(\<And> f. f \<in> set fs \<Longrightarrow> is_cnf f) \<Longrightarrow> is_cnf (form_cnf_ex fs)"
  by(induct fs, auto, auto simp: is_cnf.simps is_clause.simps)

lemma is_cnf_Conjunction[simp]: "is_cnf (Conjunction fs) = (\<forall> f \<in> set fs. is_clause f)"
  by (auto simp: is_cnf.simps)

lemma is_clause_Disjunction[simp]: "is_clause (Disjunction fs) = (\<forall> f \<in> set fs. isLit f)"
  by (auto simp: is_clause.simps)

lemma is_cnf_flatten[simp]:
  "is_cnf (flatten (\<phi> :: ('f,'v,'t) exp formula))"
  by (induct \<phi> rule: flatten.induct, auto intro!: is_cnf_form_all is_cnf_form_cnf_ex)


subsection {* Valuation *}

context fixes \<alpha> :: "'assignment" begin

  fun satisfies :: "'atom formula \<Rightarrow> bool" where
    "satisfies (Atom a) \<longleftrightarrow> eval_true \<alpha> a"
  | "satisfies (NegAtom a) \<longleftrightarrow> \<not> eval_true \<alpha> a"
  | "satisfies (Conjunction \<phi>s) \<longleftrightarrow> (\<forall>\<phi> \<in> set \<phi>s. satisfies \<phi>)"
  | "satisfies (Disjunction \<phi>s) \<longleftrightarrow> (\<exists>\<phi> \<in> set \<phi>s. satisfies \<phi>)"

  lemma satisfiesI[intro]:
    "eval_true \<alpha> a \<Longrightarrow> satisfies (Atom a)"
    "\<not> eval_true \<alpha> a \<Longrightarrow> satisfies (NegAtom a)"
    "(\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> satisfies \<phi>) \<Longrightarrow> satisfies (Conjunction \<phi>s)"
    "\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> satisfies \<phi> \<Longrightarrow> satisfies (Disjunction \<phi>s)" by auto

  lemma form_True[simp]: "satisfies True\<^sub>f" by auto

  lemma satisfies_NegAtom[simp]: "satisfies (NegAtom a) \<longleftrightarrow> \<not> satisfies (Atom a)" by simp

  lemma satisfies_and[simp]: "satisfies (\<phi> \<and>\<^sub>f \<psi>) \<longleftrightarrow> satisfies \<phi> \<and> satisfies \<psi>"
    by (cases \<phi>; cases \<psi>, auto simp: form_and.simps)

  lemma satisfies_or[simp]: "satisfies (\<phi> \<or>\<^sub>f \<psi>) \<longleftrightarrow> satisfies \<phi> \<or> satisfies \<psi>"
    by (cases \<phi>; cases \<psi>, auto simp: form_or.simps)

  lemma satisfies_all[simp]:
    "satisfies (form_all \<phi>s) \<longleftrightarrow> (\<forall>\<phi> \<in> set \<phi>s. satisfies \<phi>)" by (induct \<phi>s, auto)

  lemma satisfies_ex[simp]:
    "satisfies (form_ex \<phi>s) \<longleftrightarrow> (\<exists>\<phi> \<in> set \<phi>s. satisfies \<phi>)" by (induct \<phi>s, auto)
end

notation satisfies (infix "\<Turnstile>" 40)


subsection {* Language *}

definition Language where "Language \<phi> = {\<alpha>. \<alpha> \<Turnstile> \<phi>}"
notation Language ("\<L>") (* this allows the notation available in sublocales *)

lemma Language_simps[simp]:
  "\<L> False\<^sub>f = {}"
  "\<L> True\<^sub>f = UNIV"
  "\<L> (\<phi> \<and>\<^sub>f \<psi>) = \<L> \<phi> \<inter> \<L> \<psi>"
  "\<L> (\<phi> \<or>\<^sub>f \<psi>) = \<L> \<phi> \<union> \<L> \<psi>"
  "\<L> (NegAtom a) = UNIV - \<L> (Atom a)"
  "\<L> (Conjunction (\<phi>#\<phi>s)) = \<L> \<phi> \<inter> \<L> (Conjunction \<phi>s)"
  "\<L> (Disjunction (\<phi>#\<phi>s)) = \<L> \<phi> \<union> \<L> (Disjunction \<phi>s)" by (auto simp: Language_def)

lemma Language_Conjunction[simp]: "\<L> (Conjunction \<phi>s) = (\<Inter>\<phi> \<in> set \<phi>s. \<L> \<phi>)" by (induct \<phi>s, auto)

lemma Language_Disjunction[simp]: "\<L> (Disjunction \<phi>s) = (\<Union>\<phi> \<in> set \<phi>s. \<L> \<phi>)" by (induct \<phi>s, auto)

lemma Language_not[simp]: "\<L> (\<not>\<^sub>f \<phi>) = UNIV - \<L> \<phi>" by(induct \<phi>, auto)

lemma Language_all[simp]: "\<L> (form_all \<phi>s) = (\<Inter>\<phi> \<in> set \<phi>s. \<L> \<phi>)" by (induct \<phi>s, auto)

lemma Language_ex[simp]: "\<L> (form_ex \<phi>s) = (\<Union>\<phi> \<in> set \<phi>s. \<L> \<phi>)" by (induct \<phi>s, auto)

lemma all_conjuncts[simplified,simp]: "\<L> (form_all (conjuncts \<phi>)) = \<L> \<phi>" by (cases \<phi>, auto) 
lemma formula_conjuncts[simplified,simp]: "(\<forall>\<phi>' \<in> set (conjuncts \<phi>). formula \<phi>') = formula \<phi>" by (cases \<phi>, auto)

lemma ex_disjuncts[simplified,simp]: "\<L> (form_ex (disjuncts \<phi>)) = \<L> \<phi>" by (cases \<phi>, auto) 
lemma formula_disjuncts[simplified,simp]: "(\<forall>\<phi>' \<in> set (disjuncts \<phi>). formula \<phi>') = formula \<phi>" by (cases \<phi>, auto)

lemma Language_cnf_form_or[simp]: "\<L> (\<phi> \<or>\<^sub>c\<^sub>f \<psi>) = \<L> \<phi> \<union> \<L> \<psi>"
  by (cases \<phi>; cases \<psi>, auto)

lemma Language_cnf_ex[simp]: "\<L> (form_cnf_ex \<phi>s) = (\<Union>\<phi> \<in> set \<phi>s. \<L> \<phi>)" by (induct \<phi>s, auto)

lemma Langage_flatten[simp]: "\<L> (flatten \<phi>) = \<L> \<phi>"
  by (induct \<phi> rule: flatten.induct, auto simp: Let_def)

lemma satisfies_Language: "\<alpha> \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<in> \<L> \<phi>" by (auto simp: Language_def)

abbreviation equivalent where "equivalent \<phi> \<psi> \<equiv> (\<And>\<alpha>. assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> \<psi>)"
notation equivalent (infix "\<Longleftrightarrow>\<^sub>f" 40)

lemma Conjunction_Cons[simp]: "Conjunction (\<phi>#\<phi>s) \<Longleftrightarrow>\<^sub>f \<phi> \<and>\<^sub>f Conjunction \<phi>s" by auto

definition implies :: "'atom formula \<Rightarrow> 'atom formula \<Rightarrow> bool" where
  "implies \<phi> \<psi> \<equiv> \<forall>\<alpha>. assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> \<phi> \<longrightarrow> \<alpha> \<Turnstile> \<psi>"

notation implies (infix "\<Longrightarrow>\<^sub>f" 41)

lemma impliesI[intro]:
  assumes "\<And> \<alpha>. assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>"
  shows "\<phi> \<Longrightarrow>\<^sub>f \<psi>" using assms by (auto simp: implies_def)

lemma impliesD[dest]:
  assumes "\<phi> \<Longrightarrow>\<^sub>f \<psi>"
  shows "assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>"
  using assms by (auto simp: implies_def)

lemma impliesE[elim]:
  assumes "\<phi> \<Longrightarrow>\<^sub>f \<psi>" "(\<And>\<alpha>. assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<psi>) \<Longrightarrow> P"
  shows P using assms by (auto simp: implies_def)

(* Blocking the simplification rules for atoms... a workaround to avoid unfolding eval_true,
   which will break instantiation. *)
declare satisfies.simps(1,2)[simp del]

definition valid ("\<Turnstile>\<^sub>f") where
  "(\<Turnstile>\<^sub>f \<phi>) = (\<forall> \<alpha>. assignment \<alpha> \<longrightarrow> \<alpha> \<Turnstile> \<phi>)"  

lemma validI[intro]:
  assumes "\<And> \<alpha>. assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi>"
  shows "\<Turnstile>\<^sub>f \<phi>" using assms by (auto simp: valid_def)

lemma validD[dest]:
  assumes "\<Turnstile>\<^sub>f \<phi>"
  shows "assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi>"
  using assms by (auto simp: valid_def)
end


locale shows_formula = fixes 
  shows_atom :: "'a \<Rightarrow> shows" 
begin
fun shows_formula :: "'a formula \<Rightarrow> shows" where
  "shows_formula (Atom a) = shows_atom a" 
| "shows_formula (NegAtom a) = shows ''! ('' o shows_atom a o shows '')''" 
| "shows_formula (Conjunction fs) = (let ss = map shows_formula fs in if length ss = 1 then hd ss else
     (if ss = [] then id else shows_paren) (shows_list_gen id ''True'' [] '' && '' [] ss))" 
| "shows_formula (Disjunction fs) = (let ss = map shows_formula fs in if length ss = 1 then hd ss else
     (if ss = [] then id else shows_paren) (shows_list_gen id ''False'' [] '' || '' [] ss))" 
end




subsection {* Formulas over terms *}

fun vars_formula :: "('f,'v) term formula \<Rightarrow> 'v set"
where "vars_formula (Atom a) = vars_term a"
    | "vars_formula (NegAtom a) = vars_term a"
    | "vars_formula (Conjunction \<phi>s) = (\<Union>(vars_formula ` set \<phi>s))"
    | "vars_formula (Disjunction \<phi>s) = (\<Union>(vars_formula ` set \<phi>s))"

fun in_vars_formula :: "'v \<Rightarrow> ('f,'v) term formula \<Rightarrow> bool"
where "in_vars_formula x (Atom a) = (x \<in> set (vars_term_list a))"
    | "in_vars_formula x (NegAtom a) = (x \<in> set (vars_term_list a))"
    | "in_vars_formula x (Conjunction \<phi>s) = (\<exists> \<phi> \<in> set \<phi>s. in_vars_formula x \<phi>)" 
    | "in_vars_formula x (Disjunction \<phi>s) = (\<exists> \<phi> \<in> set \<phi>s. in_vars_formula x \<phi>)" 

lemma in_vars_formula[code_unfold]: "(x \<in> vars_formula \<phi>) = in_vars_formula x \<phi>" 
  by (induct \<phi>, auto)


abbreviation rename_vars_formula :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> ('f,'w,'t) exp formula"
where "rename_vars_formula r cls \<equiv> map_formula (rename_vars r) cls"

adhoc_overloading rename_vars rename_vars_formula

lemma rename_vars_formula_comp[simp]:
  "rename_vars f (rename_vars g (\<phi>::('f,'v,'t) exp formula)) = rename_vars (f \<circ> g) \<phi>"
   by (cases \<phi>, auto simp: formula.map_comp o_def)

record ('f,'t,'d) "logic" =
  type_of_fun :: "'f \<Rightarrow> 't list \<times> 't"
  Values_of_type :: "'t \<Rightarrow> 'd set"
  Bool_types :: "'t set"
  to_bool :: "'d \<Rightarrow> bool"
  "interpret" :: "'f \<Rightarrow> 'd list \<Rightarrow> 'd"

locale prelogic =
  fixes \<Lambda> :: "('f,'t,'d) logic"
begin

  abbreviation type_of_fun where "type_of_fun \<equiv> logic.type_of_fun \<Lambda>"
  abbreviation Values_of_type where "Values_of_type \<equiv> logic.Values_of_type \<Lambda>"
  abbreviation Bool_types where "Bool_types \<equiv> logic.Bool_types \<Lambda>"
  abbreviation to_bool where "to_bool \<equiv> logic.to_bool \<Lambda>"
  abbreviation "interpret" where "interpret \<equiv> logic.interpret \<Lambda>"

  sublocale pre_sorted_algebra "interpret" "type_of_fun" "Values_of_type".


  definition "is_bool e \<equiv> (\<exists>b \<in> Bool_types. e :\<^sub>f b) \<and> \<not> is_Var e"

  lemma is_bool_simps[simp]: "is_bool (Var x) = False"
   "is_bool (Fun f es) = (\<exists>b \<in> Bool_types. Fun f es :\<^sub>f b)" by (auto simp: is_bool_def)


  abbreviation "BOOL_EXP \<equiv> Collect is_bool"

  definition "eval_true \<alpha> t \<equiv> to_bool (\<lbrakk>t\<rbrakk>\<alpha>)"

  sublocale formula is_bool assignment eval_true .

  lemma satisfies_atom[simp]: "\<alpha> \<Turnstile> Atom e \<longleftrightarrow> to_bool (\<lbrakk>e\<rbrakk>\<alpha>)" by (auto simp: eval_true_def satisfies.simps(1))
  lemma Language_atom[simp]: "\<L> (Atom e) = { \<alpha>. to_bool (\<lbrakk>e\<rbrakk>\<alpha>) }" by (auto simp: Language_def)

  lemma is_bool_imp_expression[simp]: "is_bool e \<Longrightarrow> expression e" unfolding is_bool_def by auto

  lemma formula_rename_vars[simp]: "formula (rename_vars r e) = formula e"
  proof (induct e)
    case (Atom x)
    then show ?case by (cases x, auto)
  next
    case (NegAtom x)
    then show ?case by (cases x, auto)
  qed auto

  lemma formula_rename_vars_map[simp]:
    "list_all formula (map (rename_vars r) l) = list_all formula l"
    using formula_rename_vars unfolding list_all_def by auto

  lemma eval_with_fresh_var[simp]: "xx \<notin> vars_term e \<Longrightarrow> \<lbrakk>e\<rbrakk> (\<alpha>(xx:=a)) = \<lbrakk>e\<rbrakk> \<alpha>" by auto

  lemma satisfies_with_fresh_var[simp]:
    assumes fv: "xx \<notin> vars_formula \<phi>"
    shows "\<alpha>(xx:=a) \<Turnstile> \<phi> \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>" by (insert fv, induct \<phi>, auto)

  lemma satisfies_rename_vars:
    assumes r: "\<And> x t. \<alpha> (x,t) = \<beta> (r x, t)"
    shows "\<beta> \<Turnstile> (rename_vars r \<phi>) \<longleftrightarrow> \<alpha> \<Turnstile> \<phi>"
    by (induct \<phi>, auto simp: eval_rename_vars[of \<alpha>, OF r])

  definition satisfiable where "satisfiable \<phi> = (\<exists> \<alpha>. assignment \<alpha> \<and> \<alpha> \<Turnstile> \<phi>)" 

end

locale logic = prelogic + sorted_algebra "interpret" "type_of_fun" "Values_of_type"


subsection {* Checking validity *}

type_synonym ('h,'f,'v,'t) logic_checker = 
  "'h \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> shows check"

datatype 'h hint = Default | Base 'h | Conjunction "('h hint list)"

instantiation hint :: (default) default
begin
  definition "default \<equiv> Default"
  instance ..
end

hide_const(open) Default Base Conjunction

fun is_Atom where "is_Atom (Atom _) = True" | "is_Atom _ = False" 
fun get_Atom where "get_Atom (Atom a) = a" | "get_Atom (NegAtom a) = a" 
  
definition trivial_checker where 
  "trivial_checker lits = (case partition is_Atom lits of
     (as, nas) \<Rightarrow> let pos = map get_Atom as; neg = map get_Atom nas
    in (\<exists> a \<in> set pos. a \<in> set neg))" 
  
definition trivial_clause_checker where "trivial_clause_checker f = (case f of Disjunction lits \<Rightarrow> trivial_checker lits)" 

lemma trivial_checker: assumes "\<And> l. l \<in> set lits \<Longrightarrow> isLit l" and "trivial_checker lits"
  shows "\<exists> a. Atom a \<in> set lits \<and> NegAtom a \<in> set lits"
proof -
  obtain as nas where part: "partition is_Atom lits = (as,nas)" by force
  from assms[unfolded trivial_checker_def part split, simplified]
  obtain a where a: "a \<in> set as" "get_Atom a \<in> get_Atom ` set nas" by auto
  from partition_filter_conv[of is_Atom lits, unfolded part]
  have id: "as = filter is_Atom lits" "nas = filter (\<lambda> x. \<not> is_Atom x) lits" by (auto simp: o_def)
  from a(1)[unfolded id] have "is_Atom a" and mem: "a \<in> set lits" by auto
  then obtain aa where aa: "a = Atom aa" by (cases a, auto)
  from a(2)[unfolded aa] obtain na where na: "get_Atom na = aa" "na \<in> set nas" by auto
  from na(2)[unfolded id] assms(1) have "\<not> is_Atom na" "isLit na" "na \<in> set lits" by auto
  with na have "NegAtom aa \<in> set lits" by (cases na, auto)
  with mem aa show ?thesis by auto
qed

fun get_disjunctions where
  "get_disjunctions (Disjunction phis) = concat (map get_disjunctions phis)" 
| "get_disjunctions x = [x]" 
      
fun get_conjunctions where
  "get_conjunctions (Conjunction phis) = concat (map get_conjunctions phis)" 
| "get_conjunctions x = [x]" 

definition check_trivial_implication where
  "check_trivial_implication \<phi> \<psi> = 
      (\<forall> phi \<in> set (get_disjunctions \<phi>). let c_phis = get_conjunctions phi in  
       \<exists> psi \<in> set (get_disjunctions \<psi>). \<forall> c_psi \<in> set (get_conjunctions psi). c_psi \<in> set c_phis)" 

context prelogic
begin  
lemma get_disjunctions: "(\<alpha> \<Turnstile> form_ex (get_disjunctions \<phi>)) = (\<alpha> \<Turnstile> \<phi>)" 
  by (induct \<phi>, auto)
  
lemma get_conjunctions: "(\<alpha> \<Turnstile> form_all (get_conjunctions \<phi>)) = (\<alpha> \<Turnstile> \<phi>)" 
  by (induct \<phi>, auto)
    
lemma check_trivial_implication: assumes "check_trivial_implication \<phi> \<psi>"
  shows "\<phi> \<Longrightarrow>\<^sub>f \<psi>" 
proof
  fix \<alpha>
  assume "\<alpha> \<Turnstile> \<phi>" 
  hence "\<alpha> \<Turnstile> form_ex (get_disjunctions \<phi>)" unfolding get_disjunctions .
  then obtain phi where phi: "phi \<in> set (get_disjunctions \<phi>)" and sat: "\<alpha> \<Turnstile> phi" by auto
  from assms[unfolded check_trivial_implication_def Let_def, rule_format, OF phi]
  obtain psi where psi: "psi \<in> set (get_disjunctions \<psi>)"
    and conj: "\<And> c_psi. c_psi\<in>set (get_conjunctions psi) \<Longrightarrow> c_psi \<in> set (get_conjunctions phi)" by auto
  {
    fix c_psi
    assume "c_psi \<in> set (get_conjunctions psi)" 
    hence mem: "c_psi \<in> set (get_conjunctions phi)" using conj by auto
    from sat[folded get_conjunctions[of _ phi]] mem have "\<alpha> \<Turnstile> c_psi" by auto
  }
  hence "\<alpha> \<Turnstile> psi" using get_conjunctions[of _ psi] by auto
  thus "\<alpha> \<Turnstile> \<psi>" using psi get_disjunctions[of _ \<psi>] by auto
qed
  
end
    
locale pre_logic_checker = prelogic \<Lambda> + shows_formula shows_atom
  for \<Lambda> :: "('f::show,'t::show,'d) logic"
  and logic_checker :: "('h::{default,show},'f,'v::show,'t) logic_checker"
  and shows_atom :: "('f,'v,'t)exp \<Rightarrow> shows"
begin

lemma trivial_clause_checker: assumes "is_clause \<phi>" and "trivial_clause_checker \<phi>" shows "\<alpha> \<Turnstile> \<phi>"
proof - 
  from assms obtain lits where phi: "\<phi> = Disjunction lits" and lits: "\<And> l. l \<in> set lits \<Longrightarrow> isLit l" 
    using is_clause.cases by blast
  from assms[unfolded trivial_clause_checker_def phi] have triv: "trivial_checker lits" by auto
  from trivial_checker[OF lits this] obtain a where a: "Atom a \<in> set lits" "NegAtom a \<in> set lits" by auto
  from split_list[OF a(1)] obtain l1 l2 where lits: "lits = l1 @ Atom a # l2" by auto
  from a(2) lits have "NegAtom a \<in> set (l1 @ l2)" by auto
  from split_list[OF this] obtain l3 l4 where l12: "l1 @ l2 = l3 @ NegAtom a # l4" by auto
  have "set lits = insert (Atom a) (set (l1 @ l2))" unfolding lits by auto
  also have "\<dots> = insert (Atom a) (insert (NegAtom a) (set l3 \<union> set l4))" unfolding l12 by auto
  finally show ?thesis unfolding phi by auto
qed

definition "check_valid_formula \<phi> \<equiv>
  do {case flatten \<phi> of Conjunction \<phi>s \<Rightarrow> check_allm (
     \<lambda> \<phi>. try (check (trivial_clause_checker \<phi>) ''trivial clause checker failed'') catch (\<lambda> _. logic_checker default \<phi>)) \<phi>s}
  <+? (\<lambda>s. shows ''problem in checking validity of formula '' o shows_formula \<phi> o shows_nl o s)"

fun check_formula_hint and check_formula_hint2
where
  "check_formula_hint hint.Default \<phi> = check_valid_formula \<phi>"
| "check_formula_hint (hint.Base h) \<phi> = do {
    check (is_clause \<phi>) (shows ''only clauses may be passed to logic checker in check_formula_hint'');
    logic_checker h \<phi> <+? (\<lambda> s. shows ''problem in checking '' o shows_formula \<phi> o shows_nl o s)}"
| "check_formula_hint (hint.Conjunction hints) \<phi> =
    ( case \<phi> of formula.Conjunction \<phi>s \<Rightarrow>
      check_formula_hint2 hints \<phi>s
      | _ \<Rightarrow> error (shows ''Conjunction hint for non Conjunction '' o shows_formula \<phi>)
    )"
| "check_formula_hint2 [] [] = succeed"
| "check_formula_hint2 (h#hs) (\<phi>#\<phi>s) = check_formula_hint h \<phi> \<then> check_formula_hint2 hs \<phi>s"
| "check_formula_hint2 _ _ = error (shows ''Length mismatch in Conjunction hints'')"

lemma isOK_check_formula_hint2:
  assumes "isOK (check_formula_hint2 hs \<phi>s)" and "\<phi> \<in> set \<phi>s"
  shows "\<exists>h \<in> set hs. isOK (check_formula_hint h \<phi>)"
proof (insert assms, induct \<phi>s arbitrary: hs)
  case IH: (Cons a \<phi>s) then show ?case by(cases hs, auto)
qed auto
end

locale logic_checker =
  logic \<Lambda> + pre_logic_checker \<Lambda> logic_checker shows_atom
  for \<Lambda> :: "('f::show,'t::show,'d) logic"
  and logic_checker :: "('h::{default,show},'f,'v::show,'t) logic_checker" and shows_atom +
  assumes logic_checker:
   "\<And>h \<alpha> \<phi>. isOK(logic_checker h \<phi>) \<Longrightarrow> formula \<phi> \<Longrightarrow> is_clause \<phi> \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<alpha> \<Turnstile> \<phi>"
begin

lemma check_valid_formula:
  assumes ok: "isOK(check_valid_formula \<phi>)" and \<phi>: "formula \<phi>" and \<alpha>: "assignment \<alpha>"
  shows "\<alpha> \<Turnstile> \<phi>"
proof-
  from \<phi> have form: "formula (flatten \<phi>)" by auto
  from is_cnf_flatten[of \<phi>] have "is_cnf (flatten \<phi>)" .
  then obtain \<phi>s where *: "flatten \<phi> = Conjunction \<phi>s" "\<And> \<phi>\<^sub>i. \<phi>\<^sub>i \<in> set \<phi>s \<Longrightarrow> is_clause \<phi>\<^sub>i" by (cases, auto)
  from form have form: "\<And>\<psi>. \<psi> \<in> set \<phi>s \<Longrightarrow> formula \<psi>" by (auto simp: *)
  {
    fix \<psi>
    assume **: "\<psi> \<in> set \<phi>s" 
    note form = form[OF **]
    note clause = *(2)[OF **]
    from ok[unfolded check_valid_formula_def *, simplified] **
    consider (Triv) "trivial_clause_checker \<psi>" | (lc) "isOK(logic_checker default \<psi>)" by auto
    then have "\<alpha> \<Turnstile> \<psi>" 
    proof (cases)
      case lc
      from logic_checker[OF lc form clause \<alpha>] show ?thesis .
    next
      case Triv
      from trivial_clause_checker[OF clause this] show ?thesis . 
    qed
  }
  hence "\<alpha> \<Turnstile> (form_all \<phi>s)" by auto
  also have "... \<Longleftrightarrow>\<^sub>f flatten \<phi>" by (simp add: *)
  finally show ?thesis using \<alpha> by (simp add: satisfies_Language)
qed

lemma check_valid_formula_imp:
  assumes ok: "isOK (check_valid_formula (\<phi> \<longrightarrow>\<^sub>f \<psi>))" and form: "formula \<phi>" "formula \<psi>" shows "\<phi> \<Longrightarrow>\<^sub>f \<psi>"
  using check_valid_formula[OF ok] form by (auto simp: satisfies_Language)

lemma check_formula_hint:
  assumes "isOK(check_formula_hint h \<phi>)" and "formula \<phi>" and \<alpha>: "assignment \<alpha>"
  shows "\<alpha> \<Turnstile> \<phi>"
proof-
  have "isOK(check_formula_hint h \<phi>) \<Longrightarrow> formula \<phi> \<Longrightarrow> \<alpha> \<Turnstile> \<phi>"
   and "isOK (check_formula_hint2 hs \<phi>s) \<Longrightarrow> (\<And>\<phi>. \<phi> \<in> set \<phi>s \<Longrightarrow> formula \<phi>) \<Longrightarrow> \<alpha> \<in> (\<Inter> \<phi> \<in> set \<phi>s. \<L> \<phi>)"
   for h \<phi> hs \<phi>s
  proof (induct h \<phi> and hs \<phi>s rule:check_formula_hint_check_formula_hint2.induct)
    case (1 \<phi>) with check_valid_formula \<alpha> show ?case by auto
  next
    case (3 hints \<phi>) then show ?case by (cases \<phi>, auto simp: satisfies_Language)
  qed (auto dest: logic_checker[OF _ _ _ \<alpha>] simp: satisfies_Language)
  from this(1) assms show ?thesis by auto
qed

lemma check_formula_hint_imp:
  assumes ok: "isOK (check_formula_hint h (\<phi> \<longrightarrow>\<^sub>f \<psi>))" and form: "formula \<phi>" "formula \<psi>" shows "\<phi> \<Longrightarrow>\<^sub>f \<psi>"
  using check_formula_hint[OF ok] form by (auto simp: satisfies_Language)
end

type_synonym ('f,'v,'t) definability_checker = "'v \<Rightarrow> 't \<Rightarrow> ('f,'v,'t) exp formula \<Rightarrow> shows check"

locale pre_definability_checker = prelogic \<Lambda>
  for \<Lambda> :: "('f::show,'t::show,'d) logic"
  and definability_checker :: "('f,'v,'t) definability_checker"

locale definability_checker = pre_definability_checker +
  assumes definability_checker:
    "\<And>x ty \<phi>. isOK (definability_checker x ty \<phi>) \<Longrightarrow> assignment \<alpha> \<Longrightarrow> \<exists>a \<in> Values_of_type ty. \<alpha>((x,ty):=a) \<Turnstile> \<phi>"

lemmas [code] =
  prelogic.is_bool_def
  pre_sorted_algebra.has_type.simps
  formula.formula.simps
  shows_formula.shows_formula.simps
  pre_logic_checker.check_valid_formula_def
  pre_logic_checker.check_formula_hint.simps
  pre_logic_checker.check_formula_hint2.simps

(* TODO: Move *)
fun or_ok :: "'a check \<Rightarrow> 'a check \<Rightarrow> 'a check" where
"or_ok (Inl a) b = b" |
"or_ok (Inr a) b = Inr a" 

lemma or_is_or: "isOK (or_ok a b) = isOK a \<or> isOK b" using or_ok.elims by blast

end
