theory Linear_Integer_Arithmetic
imports
  Formula
  "$AFP/Show/Show_Instances"
  "$AFP/Deriving/Comparator_Generator/Compare_Generator"
  "$AFP/Certification_Monads/Check_Monad"
  Non_Inf_Orders
begin


section {* Definition of LIA *}

locale LIA
begin
datatype ty = BoolT | IntegerT
datatype sig = Less | Lesseq | Sum | Const (the_Const: int) | Cmult int 
datatype sig' = Less' | Lesseq' | Sum' | Const' (the_Const: int) | Cmult' int | Equals'
datatype "value" = Integer (the_int: int) | Boolean (the_bool: bool)

end

type_synonym 'v LIA_exp = "(LIA.sig,'v,LIA.ty) exp"
type_synonym 'v LIA_exp' = "(LIA.sig','v,LIA.ty) exp"
type_synonym 'v LIA_formula = "(LIA.sig,'v,LIA.ty) exp formula"
type_synonym 'v LIA_formula' = "(LIA.sig','v,LIA.ty) exp formula"

interpretation LIA_logic: LIA.

derive "show" LIA.sig
derive "show" LIA.ty
derive compare_order LIA.value

context LIA begin

fun show_LIA_exp where
  "show_LIA_exp (Var (x,ty)) = shows x"
| "show_LIA_exp (Fun (Const n) []) = shows n"
| "show_LIA_exp (Fun (Cmult n) [Var (x,ty)]) = shows n o shows '' * '' o shows_paren (shows x)"
| "show_LIA_exp (Fun Less [s,t]) = show_LIA_exp s o shows '' < '' o show_LIA_exp t"
| "show_LIA_exp (Fun Lesseq [s,t]) = show_LIA_exp s o shows '' <= '' o show_LIA_exp t"
| "show_LIA_exp (Fun Sum [s,t]) = show_LIA_exp s o shows '' + '' o show_LIA_exp t"
| "show_LIA_exp _ = shows ''<< unexpected LIA expression >>''"

definition "show_LIA_clause = shows_list_gen show_LIA_exp ''FALSE'' [] '' || '' []" 

definition "show_LIA_formula = shows_list_gen (shows_paren o show_LIA_clause) ''TRUE'' [] '' && '' []" 


fun LIA_type_of_fun :: "sig \<Rightarrow> ty list \<times> ty" where
  "LIA_type_of_fun Less = ([IntegerT,IntegerT], BoolT)"
| "LIA_type_of_fun Lesseq = ([IntegerT,IntegerT], BoolT)"
| "LIA_type_of_fun Sum = ([IntegerT,IntegerT], IntegerT)"
| "LIA_type_of_fun (Const _) = ([], IntegerT)"
| "LIA_type_of_fun (Cmult _) = ([IntegerT], IntegerT)" (* only admit multiplication with constant *)

fun LIA_Values_of_type :: "ty \<Rightarrow> value set" where
  "LIA_Values_of_type BoolT = range Boolean"
| "LIA_Values_of_type IntegerT = range Integer"

fun "LIA_interpret" :: "sig \<Rightarrow> value list \<Rightarrow> value" where
  "LIA_interpret Less [x, y] = Boolean (the_int x < the_int y)"
| "LIA_interpret Lesseq [x, y] = Boolean (the_int x \<le> the_int y)"
| "LIA_interpret Sum [x, y] = Integer (the_int x + the_int y)"
| "LIA_interpret (Const x) [] = Integer x"
| "LIA_interpret (Cmult x) [y] = Integer (x * the_int y)"
| "LIA_interpret _ _ = undefined"

definition "\<Lambda> = \<lparr>
  type_of_fun = LIA_type_of_fun,
  Values_of_type = LIA_Values_of_type,
  Bool_types = {BoolT},
  to_bool = the_bool,
  interpret = LIA_interpret
\<rparr>"

sublocale prelogic \<Lambda>.

lemma LIA_fold: "LIA_type_of_fun = type_of_fun"
  "LIA_Values_of_type = Values_of_type"
  "the_bool = to_bool"
  "LIA_interpret = interpret" by (auto simp: \<Lambda>_def)

lemma Bool_types[simp]: "Bool_types = {BoolT}" by (auto simp: \<Lambda>_def)

lemmas interpret_simps[simp] = LIA_interpret.simps[unfolded LIA_fold]
   and type_of_fun_simps[simp] = LIA_type_of_fun.simps[unfolded LIA_fold]
   and Values_of_type_simps[simp] = LIA_Values_of_type.simps[unfolded LIA_fold]
   and to_bool_simp[simp] = LIA.value.sel(2)[unfolded LIA_fold]

lemma mem_Values_of_typeE[elim]:
  assumes "v \<in> Values_of_type ty"
      and "\<And>i. ty = IntegerT \<Longrightarrow> v = Integer i \<Longrightarrow> thesis"
      and "\<And>b. ty = BoolT \<Longrightarrow> v = Boolean b \<Longrightarrow> thesis"
  shows "thesis" using assms by (cases ty, auto)

lemma mem_Values_of_typeD[dest]:
  assumes "v \<in> Values_of_type ty"
  shows "(ty = IntegerT \<longrightarrow> (\<exists>i. v = Integer i)) \<and> (ty = BoolT \<longrightarrow> (\<exists>b. v = Boolean b))"
  by (insert assms, auto)

lemma has_type_induct[consumes 1, case_names Less Lesseq Sum Const Cmult Var]:
  assumes "e :\<^sub>f ty"
  and LessCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Less [ta,tb]) BoolT"
  and LesseqCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Lesseq [ta,tb]) BoolT"
  and SumCase: "\<And> ta tb.
       ta :\<^sub>f IntegerT \<Longrightarrow> tb :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P tb IntegerT \<Longrightarrow> P (Fun Sum [ta,tb]) IntegerT"
  and ConstCase: "\<And> x. P (Fun (Const x) []) IntegerT"
  and CmultCase: "\<And> x ta. ta :\<^sub>f IntegerT \<Longrightarrow> P ta IntegerT \<Longrightarrow> P (Fun (Cmult x) [ta]) IntegerT"
  and VarCase: "\<And> x ty. P (Var (x,ty)) ty"
  shows "P e ty"
using assms(1) proof(induct e ty rule: has_type.induct)
  case (1 x ty) with VarCase show ?case by(cases x, auto)
next
  case Fun: (2 f ls ty)
  show ?case
  proof(cases f)
    case [simp]: Less
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LessCase show ?thesis by auto
  next
    case [simp]: Lesseq
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls rule: list_3_cases, auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = BoolT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] LesseqCase show ?thesis by auto
  next
    case [simp]: Sum
    from Fun(2) obtain ta tb where [simp]: "ls = [ta,tb]" by (cases ls; cases "tl ls", auto)
    from Fun(2) have at: "ta :\<^sub>f IntegerT" and bt: "tb :\<^sub>f IntegerT" and ty: "ty = IntegerT" by auto
    with Fun.hyps[of 0] Fun.hyps[of 1] SumCase show ?thesis by auto
  next
    case [simp]: (Const x)
    with Fun have [simp]: "ls = []" by auto
    show ?thesis using Fun ConstCase by simp
  next
    case [simp]: (Cmult x)
    from Fun(2) obtain ta where [simp]: "ls = [ta]" by (cases ls, auto)
    from Fun(2) Fun.hyps[of 0] show ?thesis using CmultCase by auto
  qed
qed

sublocale logic \<Lambda>
proof
  fix f ds
  show "length ds = length (param_types f) \<Longrightarrow>
       (\<And>i. i < length ds \<Longrightarrow> ds ! i \<in> Values_of_type (param_types f ! i)) \<Longrightarrow>
       interpret f ds \<in> Values_of_type (return_type f)"
  by (cases f; induct f ds rule: LIA_interpret.induct, auto)
  show "Values_of_type ty \<noteq> {}" for ty by(cases ty, auto)
qed

lemmas [simp] = less_value_def compare_value_def lt_of_comp_def comparator_of_def

lemma less_LIA_value_simps[simp]:
 "\<And>x y. Boolean x > Boolean y \<longleftrightarrow> x > y"
 "\<And>x y. Boolean x > Integer y"
 "\<And>x y. Integer x > Integer y \<longleftrightarrow> x > y"
 "\<And>x y. Integer x > Boolean y \<longleftrightarrow> False"
proof-
  show "Boolean x > Boolean y \<longleftrightarrow> x > y" for x y by (cases x; cases y, auto)
qed auto

lemmas [simp del] = less_value_def compare_value_def lt_of_comp_def comparator_of_def

lemma LIA_value_less_eq[simp]: "(x :: value) \<ge> y \<longleftrightarrow> x = y \<or> x > y" by auto

end

instance "LIA.value" :: non_inf_quasi_order
proof(intro_classes)
  fix f :: "nat \<Rightarrow> LIA.value" and a
  assume "\<forall>i. f (Suc i) < f i"
  note chain = this[rule_format]
  have f0: "\<not> LIA.Boolean True < f 0" by (cases "f 0", auto)
  from chain[of 0] have f1: "\<not> LIA.Boolean False < f 1" by (cases "f 0"; cases "f 1", auto)
  have "\<exists>x. f (Suc (Suc n)) = LIA.Integer x" for n
  proof(induct n)
    case 0
    from f1 chain[of 1] have "f (Suc (Suc 0)) < LIA.Boolean False" by auto
    then show ?case by (cases "f (Suc (Suc 0))", auto)
  next
    case (Suc n)
    then obtain x where IH: "f (Suc (Suc n)) = LIA.Integer x" by auto
    with chain[of "Suc (Suc n)"] show ?case by (cases "f (Suc (Suc (Suc n)))", auto)
  qed
  from choice[OF allI[OF this]] obtain g where g: "\<And>i. f (Suc (Suc i)) = LIA.Integer (g i)" by force
  then have chaing: "g (Suc i) < g i" for i using chain[of "Suc (Suc i)"] by auto
  with non_inf_int_gt[unfolded non_inf_def]
  have "\<And>n. \<exists>i. \<not> n < g i" by blast
  with g have "\<exists>i. \<not> LIA.Integer n < f (Suc (Suc i))" for n by auto
  then have *: "\<exists>i. \<not> LIA.Integer n < f i" for n by auto
  show "\<exists>i. \<not> a < f i"
  proof(cases a)
    case a: (Boolean b) with f0 f1 show ?thesis by (cases b, auto)
  next
    case a: (Integer i) with * show ?thesis by auto
  qed
qed

context LIA begin

definition add :: "'v LIA_exp \<Rightarrow> 'v LIA_exp \<Rightarrow> 'v LIA_exp" where "add h t \<equiv> Fun Sum [h,t]"
abbreviation "lia_const c \<equiv> Fun (Const c) []"
abbreviation "lia_var v \<equiv> Var (v, IntegerT)"

fun of_LIA_exp' :: "'v LIA_exp' \<Rightarrow> 'v LIA_exp" where
  "of_LIA_exp' (Fun (Const' i) []) = Fun (Const i) []" 
| "of_LIA_exp' (Fun (Cmult' i) [x]) = Fun (Cmult i) [of_LIA_exp' x]" 
| "of_LIA_exp' (Fun Sum' xs) = (let ys = map of_LIA_exp' xs in if ys = [] then 
     lia_const 0 else foldl add (hd ys) (tl ys))" 
| "of_LIA_exp' (Var x) = (Var x)" 
| "of_LIA_exp' e = (Code.abort (STR ''unknown LIA' expression'') (\<lambda> _. lia_const 0))" 

fun of_LIA_atom' :: "'v LIA_exp' \<Rightarrow> 'v LIA_formula" where
  "of_LIA_atom' (Fun Less' [x,y]) = Atom (Fun Less [of_LIA_exp' x, of_LIA_exp' y])" 
| "of_LIA_atom' (Fun Lesseq' [x,y]) = Atom (Fun Lesseq [of_LIA_exp' x, of_LIA_exp' y])" 
| "of_LIA_atom' (Fun Equals' [x,y]) = (let x' = of_LIA_exp' x; y' = of_LIA_exp' y
     in Conjunction [Atom (Fun Lesseq [x', y']), Atom (Fun Lesseq [y',x'])])" 

fun of_LIA_formula' :: "'v LIA_formula' \<Rightarrow> 'v LIA_formula" where
  "of_LIA_formula' (Conjunction fs) = Conjunction (map of_LIA_formula' fs)" 
| "of_LIA_formula' (Disjunction fs) = Disjunction (map of_LIA_formula' fs)" 
| "of_LIA_formula' (Atom a) = of_LIA_atom' a" 
| "of_LIA_formula' (NegAtom a) = (\<not>\<^sub>f (of_LIA_atom' a))" 

end

subsection {* Flattened version *}

datatype 'v lin_exp = Lin_Exp (lin_const:int) (lin_list:"('v \<times> int)list") (* meaning: c + x1 * c1 + ... xn * cn \<ge> 0 *)

context LIA begin

fun shows_lin_exp :: "'v :: show lin_exp \<Rightarrow> shows" where 
  "shows_lin_exp (Lin_Exp c cxs) = (shows_list_gen (\<lambda> (v,c). if c = 1 then shows v else shows c o shows '' * '' o shows v)
     ''0'' [] '' + '' [] (filter (\<lambda> (v,c). c \<noteq> 0) cxs) o shows '' >= '' o shows (-c))"

fun add_and_mult_list :: "nat \<Rightarrow> 'v lin_exp \<Rightarrow> 'v lin_exp \<Rightarrow> 'v lin_exp" where
"add_and_mult_list c le1 le2 = Lin_Exp (lin_const le1 * (int c) + lin_const le2)
                                 (map (apsnd (op * (int c))) (lin_list le1) @ lin_list le2)"

fun interpret_lin_list :: "('v \<Rightarrow> int) \<Rightarrow> ('v\<times>int) list \<Rightarrow> int"
  where
  "interpret_lin_list f Nil = 0" |
  "interpret_lin_list f (Cons a as) = f (fst a) * (snd a) + interpret_lin_list f as"

fun interpret_lin_exp :: "('v \<Rightarrow> int) \<Rightarrow> 'v lin_exp \<Rightarrow> int"
  where "interpret_lin_exp f le = lin_const le + interpret_lin_list f (lin_list le)"

definition interpret_lin_exp_atom :: "('v \<Rightarrow> int) \<Rightarrow> 'v lin_exp \<Rightarrow> bool" where
  "interpret_lin_exp_atom f le = (interpret_lin_exp f le \<ge> 0)" 

fun flatten_exp :: "'v LIA_exp \<Rightarrow> 'v lin_exp" where
"flatten_exp (Var a) = Lin_Exp 0 [(fst a,1)]" |
"flatten_exp (Fun Less [a, b])
  = (case (flatten_exp a,flatten_exp b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc - ac - 1) (map (apsnd (op * (-1))) al @ bl))" |
"flatten_exp (Fun Lesseq [a, b])
  = (case (flatten_exp a,flatten_exp b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc - ac) (map (apsnd (op * (-1))) al @ bl))" |
"flatten_exp (Fun Sum [a, b])
  = (case (flatten_exp a,flatten_exp b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc + ac) (al @ bl))" |
"flatten_exp (Fun (Const a) [])
  = Lin_Exp a []" |
"flatten_exp (Fun (Cmult c) [a])
  = (case (flatten_exp a) of
       (Lin_Exp ac al) \<Rightarrow> Lin_Exp (ac*c) (map (apsnd (op * c)) al))"

fun normalize_sum_duplicates :: "('v \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum_duplicates [] = []"
| "normalize_sum_duplicates ((x,c) # cxs) = (case partition (\<lambda> (y,d). y = x) cxs of
     (xs, dys) \<Rightarrow> (x,c + sum_list (map snd xs)) # normalize_sum_duplicates dys)"

definition normalize_sum_0 :: "('v \<times> int)list \<Rightarrow> ('v \<times> int)list" where
  "normalize_sum_0 = filter (\<lambda> (x,c). c \<noteq> 0)"

definition normalize_sum :: "('v :: linorder \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum = normalize_sum_0 o sort o normalize_sum_duplicates"

fun compress :: "('v::linorder) lin_exp \<Rightarrow> 'v lin_exp"
where
"compress v = Lin_Exp (lin_const v) (normalize_sum (lin_list v))"

lemma partition_sum_list: "partition f xs = (yes,no) \<Longrightarrow> sum_list (map g xs) = 
  sum_list (map g yes) + (sum_list (map g no) :: 'a :: ab_group_add)"
proof (induct xs arbitrary: yes no)
  case (Cons x xs)
  obtain y n where p: "partition f xs = (y,n)" by force
  from Cons(2)[unfolded partition.simps p Let_def split]
  have id: "(yes, no) = (if f x then (x # y, n) else (y, x # n))" by auto
  from Cons(1)[OF p] have IH: "sum_list (map g xs) = sum_list (map g y) + sum_list (map g n)" .
  show ?case using id by (cases "f x", auto simp: IH ac_simps)
qed simp

lemma interpret_append[simp]:
  "interpret_lin_list f (a @ b) = interpret_lin_list f a + interpret_lin_list f b"
by(induct a,simp_all)

lemma interpret_mult[simp]:
  "interpret_lin_list f (map (apsnd (op * c)) x) = c * interpret_lin_list f x"
by(induct x,auto simp:distrib_left)

lemma add_and_mult_list_ok[simp]:
  assumes a1:"0 \<le> interpret_lin_exp f l1" and a2:"0 \<le> interpret_lin_exp f l2"
  shows "0 \<le> interpret_lin_exp f (add_and_mult_list n l1 l2)"
proof -
  have "0 \<le> int n * (lin_const l1 + interpret_lin_list f (lin_list l1))"
    using a1 by simp
  then show ?thesis using a2 by (simp add: distrib_right mult_of_nat_commute)
qed

declare add_and_mult_list.simps[simp del]

lemma flatten_exp_agrees_int:
  assumes \<alpha>: "assignment \<alpha>"
  and "e :\<^sub>f IntegerT"
  shows "\<lbrakk>e\<rbrakk>\<alpha> = Integer (interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (flatten_exp e))"
proof -
  have "e :\<^sub>f IntegerT \<Longrightarrow> ?thesis" for ty
  proof(induct e IntegerT rule:has_type_induct)
    case (Var x)
    from \<alpha> obtain i where "\<alpha> (x, IntegerT) = Integer i" by force
    then show ?case by auto
  qed(simp_all add:lin_exp.case_eq_if distrib_left)
  thus ?thesis using assms by simp
qed

lemma flatten_exp_agrees_bool:
  assumes \<alpha>: "assignment \<alpha>"
  and e: "is_bool e"
  shows "to_bool (\<lbrakk>e\<rbrakk>\<alpha>) = (interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (flatten_exp e) \<ge> 0)"
proof -
  have "e :\<^sub>f BoolT \<Longrightarrow> is_Fun e \<Longrightarrow> ?thesis"
    by (induct e BoolT rule: has_type_induct, insert flatten_exp_agrees_int[OF \<alpha>], auto simp:lin_exp.case_eq_if)
  then show ?thesis using e[unfolded is_bool_def] by auto
qed
  
lemma normalize_sum_duplicates[simp]: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum_duplicates cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence xs: "\<And> y c. (y,c) \<in> set xs \<Longrightarrow> y = x" by auto
  show ?case unfolding normalize_sum_duplicates.simps part split
    by (simp add: field_simps partition_sum_list[OF part] 2[OF part[symmetric]],
    insert xs, induct xs, (force simp: field_simps)+)
qed simp

lemma normalize_sum_0: "(\<Sum>(x, c)\<leftarrow> normalize_sum_0 cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
  unfolding normalize_sum_0_def by (induct cxs, auto)

lemma normalize_sum_duplicates_distinct: 
  "distinct (map fst (normalize_sum_duplicates cxs)) \<and> fst ` set (normalize_sum_duplicates cxs) = fst ` set cxs"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence dys: "\<And> y c. (y,c) \<in> set dys \<Longrightarrow> y \<noteq> x" by auto
  show ?case using 2[OF part[symmetric]] unfolding normalize_sum_duplicates.simps part split
    using part by auto
qed simp

lemma normalize_sum: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof -
  from normalize_sum_duplicates_distinct[of cxs]
  have dist: "distinct (normalize_sum_duplicates cxs)"
    by (simp add: distinct_map)
  show ?thesis
    unfolding normalize_sum_def o_def normalize_sum_0
    by (subst sum.distinct_set_conv_list[symmetric], simp add: dist, simp,
    subst sum.distinct_set_conv_list[OF dist], simp)
qed

lemma interpret_lin_list_connection: "(\<Sum>(x, c)\<leftarrow> cxs. c * \<alpha> x) = interpret_lin_list \<alpha> cxs"
by(induct cxs,auto)

lemma normalize_ok: "interpret_lin_list \<alpha> (normalize_sum cxs) = interpret_lin_list \<alpha> cxs"
using normalize_sum by (auto simp:interpret_lin_list_connection)

lemma compress_ok: "interpret_lin_exp \<alpha> (compress cxs) = interpret_lin_exp \<alpha> cxs"
using normalize_ok by auto

definition def_checker_main where
  "def_checker_main x ty l1 r1 l2 r2 = (is_bool (Fun Lesseq [l1, r1]) \<and>
    r1 :\<^sub>f ty \<and> is_bool (Fun Lesseq [l2, r2]) \<and> Var (x,ty) = l1 \<and> l1 = r2 \<and> l2 = r1 \<and> (x,ty) \<notin> vars_term r1)" 

definition def_checker :: "(sig, 'v :: show, ty) definability_checker"
(* unfortunately, a nice pattern matching doesn't go well for proving... *)
where "def_checker x ty \<phi> \<equiv>
 (case \<phi> of Conjunction \<phi>s \<Rightarrow>
   (case \<phi>s of [] \<Rightarrow> succeed
    | [lt1, lt2] \<Rightarrow>
     (case lt1 of Atom e1 \<Rightarrow>
       (case e1 of Fun Lesseq [l1,r1] \<Rightarrow>
         (case lt2 of Atom e2 \<Rightarrow>
           (case e2 of Fun Lesseq [l2,r2] \<Rightarrow>
              check (def_checker_main x ty l1 r1 l2 r2 \<or> def_checker_main x ty l2 r2 l1 r1)
              (shows ''Non-equation not supported'')
             | _ \<Rightarrow> error (shows ''Invalid 2nd literal'')
           )
          | _ \<Rightarrow> error (shows ''Invalid 2nd clause'')
         )
        | _ \<Rightarrow> error (shows ''Invalid 1st literal'')
        )
      | _ \<Rightarrow> error (shows ''Invalid 1st clause'')
     )
   | _ \<Rightarrow> error (shows ''Conjunction of more than two'')
  )
 | _ \<Rightarrow> error (shows ''Invalid formula''))
 <+? (\<lambda> s. shows ''error in definibility checker for '' o shows x o shows '' on formula\<newline>  '' o shows_formula.shows_formula show_LIA_exp \<phi> o shows_nl o s)"

sublocale definability_checker \<Lambda> "def_checker :: (_,'v :: show,_) definability_checker"
proof
  fix \<alpha> :: "'v \<times> ty \<Rightarrow> value" and x :: 'v and ty \<phi>
  assume ok: "isOK (def_checker x ty \<phi>)" and alpha: "assignment \<alpha>"
  note ok = ok[unfolded def_checker_def, simplified]
  show "\<exists>a \<in> Values_of_type ty. \<alpha>((x, ty) := a) \<Turnstile> \<phi>"
  proof (cases \<phi>)
    case [simp]: (Conjunction \<phi>s)
    note ok = ok[simplified]
    show ?thesis
    proof (cases \<phi>s rule: list_4_cases)
      case Nil with ok show ?thesis by (cases ty, auto)
    next
      case [simp]: (2 cl1 cl2)
      note ok = ok[simplified]
      then obtain l1 r1 where [simp]: "cl1 = Atom (Fun Lesseq [l1,r1])"
        by (split formula.splits, split term.splits, auto split: list.splits sig.splits)
      note ok = ok[simplified]
      then obtain l2 r2 where [simp]: "cl2 = Atom (Fun Lesseq [l2,r2])"
        by (split formula.splits, split term.splits, auto split: list.splits sig.splits)
      note ok = ok[simplified]
      { fix l1 r1 l2 r2
        assume "def_checker_main x ty l1 r1 l2 r2"
        note ok = this[unfolded def_checker_main_def]
        then have "is_bool (Fun Lesseq [l1, r1])" and "is_bool (Fun Lesseq [l2, r2])" by auto
        from ok have id: "l1 = Var (x,ty)" "r2 = Var (x,ty)" "l2 = r1" and fresh: "(x,ty) \<notin> vars_term r1" by auto
        from fresh have id2: "(\<lbrakk>r1\<rbrakk> (\<alpha>((x, ty) := a))) = (\<lbrakk>r1\<rbrakk> \<alpha>)" for a by simp
        from ok have "r1 :\<^sub>f ty" by auto
        with alpha have r1: "\<lbrakk>r1\<rbrakk> \<alpha> \<in> Values_of_type ty"
          using eval_types by auto
        have "(\<exists>a \<in> Values_of_type ty.
        \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun Lesseq [l1, r1]) \<and>
        \<alpha>((x, ty) := a) \<Turnstile> Atom (Fun Lesseq [l2, r2]))" 
          unfolding id using r1 by (intro bexI[of _ "\<lbrakk>r1\<rbrakk> \<alpha>"], auto simp: id2)
      } note [dest!] = this
      from ok show ?thesis by auto
    qed (insert ok, auto)
  qed (insert ok, auto)
qed
end

end