theory Quasi_Order
  imports "../Orderings/Reduction_Pair"
begin


section{* Quasi-Ordered Domains *}

context ord begin

text {* Extending @{class ord} with some more notations. *}

abbreviation less_eq_set ("{\<le>}") where "{\<le>} \<equiv> {(x,y). x \<le> y}"
abbreviation less_set ("{<}") where "{<} \<equiv> {(x,y). x < y}"
abbreviation greater_eq_set ("{\<ge>}") where "{\<ge>} \<equiv> {(x,y). x \<ge> y}"
abbreviation greater_set ("{>}") where "{>} \<equiv> {(x,y). x > y}"

definition equiv (infix "\<sim>" 50) where [iff]: "x \<sim> y \<equiv> x \<le> y \<and> y \<le> x"

lemma equiv_sym[sym]: "x \<sim> y \<Longrightarrow> y \<sim> x" by auto

definition equiv_class ("[_]\<^sub>\<sim>") where "[x]\<^sub>\<sim> \<equiv> { y. x \<sim> y }"

lemma mem_equiv_class[simp]: "y \<in> [x]\<^sub>\<sim> \<longleftrightarrow> x \<sim> y" by (auto simp: equiv_class_def)

definition "UB X \<equiv> {b. \<forall>x \<in> X. b \<ge> x}"

lemma UB_I[intro]: "(\<And>x. x \<in> X \<Longrightarrow> b \<ge> x) \<Longrightarrow> b \<in> UB X" by (auto simp: UB_def)

lemma UB_D[dest]: "b \<in> UB X \<Longrightarrow> x \<in> X \<Longrightarrow> b \<ge> x" by (auto simp: UB_def)

lemma mem_UB[simp]: "b \<in> UB X \<longleftrightarrow> (\<forall>x \<in> X. b \<ge> x)" by auto

lemma UB_cmono: assumes "X \<subseteq> Y" shows "UB X \<supseteq> UB Y" using assms by auto

definition "LUB X \<equiv> { b \<in> UB X. \<forall>c \<in> UB X. b \<le> c }"

lemma LUB_I[intro]: assumes "b \<in> UB X" "\<And>c. c \<in> UB X \<Longrightarrow> b \<le> c" shows "b \<in> LUB X"
  using assms by (auto simp: LUB_def)

lemma LUB_D[dest]: assumes "b \<in> LUB X" shows "b \<in> UB X" "c \<in> UB X \<Longrightarrow> b \<le> c"
  using assms[unfolded LUB_def] by auto

lemma LUB_mono:
  assumes XY: "X \<subseteq> Y" and bX: "bX \<in> LUB X" and bY: "bY \<in> LUB Y"
  shows "bX \<le> bY"
proof-
  have "bY \<in> UB X" using XY bY by auto
  with bX show ?thesis by auto
qed

end

text {* This is a trick to make a class available also as a locale *}

locale ord_syntax = ord less_eq less for less_eq less :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
begin

notation less_eq (infix "\<sqsubseteq>" 50)
notation less (infix "\<sqsubset>" 50)
abbreviation(input) greater_eq_syntax (infix "\<sqsupseteq>" 50) where "greater_eq_syntax \<equiv> ord.greater_eq less_eq"
abbreviation(input) greater_syntax (infix "\<sqsupset>" 50) where "greater_syntax \<equiv> ord.greater less"
abbreviation less_eq_set_syntax ("{\<sqsubseteq>}") where "{\<sqsubseteq>} \<equiv> ord.less_eq_set less_eq"
abbreviation less_set_syntax ("{\<sqsubset>}") where "{\<sqsubset>} \<equiv> ord.less_set less"
abbreviation greater_eq_set_syntax ("{\<sqsupseteq>}") where "{\<sqsupseteq>} \<equiv> ord.greater_eq_set less_eq"
abbreviation greater_set_syntax ("{\<sqsupset>}") where "{\<sqsupset>} \<equiv> ord.greater_set less"

no_notation equiv (infix "\<sim>" 50)
abbreviation equiv_syntax (infix "\<sim>" 50) where "equiv_syntax \<equiv> ord.equiv less_eq"

no_notation equiv_class ("[_]\<^sub>\<sim>")
abbreviation equiv_class_syntax ("[_]\<^sub>\<sim>") where "equiv_class_syntax \<equiv> ord.equiv_class less_eq"

abbreviation UB_syntax ("UB") where "UB \<equiv> ord.UB less_eq"

abbreviation LUB_syntax ("LUB") where "LUB \<equiv> ord.LUB less_eq"

end

class quasi_order = ord +
  assumes order_refl[iff]: "x \<le> x"
      and order_trans[trans]: "x \<le> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z"
      and less_trans[trans]: "x < y \<Longrightarrow> y < z \<Longrightarrow> x < z"
      and le_less_trans[trans]: "x \<le> y \<Longrightarrow> y < z \<Longrightarrow> x < z"
      and less_le_trans[trans]: "x < y \<Longrightarrow> y \<le> z \<Longrightarrow> x < z"
begin

text {*
 Class @{class quasi_order} is incomparable with HOL class @{class preorder}, as the latter
 fixes $<$ as the "strict part" (which is not nice, e.g., for comparing polynomials)
 while we already assume compatibility (@{thm le_less_trans}).
*}

lemma equiv_trans[trans]:
  "x \<sim> y \<Longrightarrow> y \<sim> z \<Longrightarrow> x \<sim> z"
  "x \<sim> y \<Longrightarrow> y \<le> z \<Longrightarrow> x \<le> z"
  "x \<le> y \<Longrightarrow> y \<sim> z \<Longrightarrow> x \<le> z"
  "x \<sim> y \<Longrightarrow> y < z \<Longrightarrow> x < z"
  "x < y \<Longrightarrow> y \<sim> z \<Longrightarrow> x < z"
  by (auto dest: order_trans le_less_trans less_le_trans)

lemma LUB_singleton: "LUB {x} = [x]\<^sub>\<sim>"
proof(intro equalityI subsetI; unfold mem_equiv_class)
  fix b assume "b \<in> LUB {x}" then have "b \<in> UB {x}" "\<And>c. c \<in> UB {x} \<Longrightarrow> b \<le> c" by auto
  then have "b \<ge> x" "\<And>c. c \<ge> x \<Longrightarrow> b \<le> c" by auto
  then show "x \<sim> b" by auto
next
  fix b assume xb: "x \<sim> b" then show "b \<in> LUB {x}" by(intro LUB_I, auto dest:order_trans)
qed

lemma LUB_equiv: assumes bLUB: "b \<in> LUB X" shows "LUB X = [b]\<^sub>\<sim>"
proof(intro equalityI subsetI, unfold mem_equiv_class)
  from bLUB have bUB: "b \<in> UB X" and leastb: "\<And>x. x \<in> UB X \<Longrightarrow> b \<le> x" by auto
  { fix c assume "c \<in> LUB X"
    then have cUB: "c \<in> UB X" and leastc: "\<And>b. b \<in> UB X \<Longrightarrow> c \<le> b" by auto
    from leastb[OF cUB] leastc[OF bUB] show "b \<sim> c" by auto
  }
  { fix c assume bc: "b \<sim> c"
    show "c \<in> LUB X"
    proof(intro LUB_I UB_I)
      fix x assume "x \<in> X"
      with bUB have "x \<le> b" by auto
      with bc show "x \<le> c" by (auto dest: order_trans)
    next
      fix x assume "x \<in> UB X"
      from leastb[OF this] bc show "c \<le> x" by (auto dest: order_trans)
    qed
  }
qed

sublocale order_pair "{>}" "{\<ge>}"
  apply (unfold_locales, fold transp_trans)
  using le_less_trans less_le_trans
  by (auto intro: refl_onI transpI[OF less_trans] transpI[OF order_trans])

end

subsection {* Weakly ordered domain: where antisymmetry holds *}
class weak_order = quasi_order + assumes antisym[dest]: "a \<le> b \<Longrightarrow> b \<le> a \<Longrightarrow> a = b"
begin

lemma equiv_is_eq[simp]: "op \<sim> = op =" by (intro ext, auto)

lemma equiv_class_singleton[simp]: "[s]\<^sub>\<sim> = {s}" by auto

end

subclass(in order) weak_order by(unfold_locales, auto)


subsection {* Well-founded quasi-ordered domain *}

class wf_order = quasi_order +
  assumes less_induct[case_names less]: "\<And>P a. (\<And>x. (\<And>y. y < x \<Longrightarrow> P y) \<Longrightarrow> P x) \<Longrightarrow> P a"
begin

lemma wf: "wf {<}" unfolding wf_def by(auto intro:less_induct)

sublocale SN_order_pair "{>}" "{\<ge>}"
  by (unfold_locales, auto simp: SN_iff_wf converse_def conversep.simps intro: wf)

end


text {* @{class wellorder} is a @{class wf_order}. *}

subclass(in wellorder) wf_order by (unfold_locales; (fact less_induct | simp))

text {* @{type nat} is a @{class wf_order}. *}
instance nat :: wf_order ..



subsection {* Lattice-like structures *}

class quasi_order_sup = quasi_order + sup +
  assumes sup_ge1 [simp]: "x \<le> sup x y"
      and sup_ge2 [simp]: "y \<le> sup x y"
begin

lemma sup_trans1: assumes "x \<le> y" shows "x \<le> sup y z"
proof-
  note assms also note sup_ge1 finally show ?thesis by auto
qed

lemma sup_trans2: assumes "x \<le> z" shows "x \<le> sup y z"
proof-
  note assms also note sup_ge2 finally show ?thesis by auto
qed

lemma finite_set_has_UB: "finite X \<Longrightarrow> \<exists>b. b \<in> UB X"
proof(unfold UB_def, induct X rule:finite_induct)
  case empty show ?case by auto
next
  case (insert x X)
  then obtain b where *: "\<And>x. x \<in> X \<Longrightarrow> x \<le> b" by auto
  with sup_trans1 have "y \<in> insert x X \<Longrightarrow> y \<le> sup b x" for y by (elim insertE,auto)
  then show ?case by auto
qed

lemma finite_set_is_bounded:
  assumes "finite X" and "x \<in> X" shows "x \<le> (SOME b. b \<in> UB X)"
  using finite_set_has_UB[OF assms(1), folded some_eq_ex] assms(2) by auto

end

class quasi_semilattice_sup = quasi_order_sup +
  assumes sup_least: "y \<le> x \<Longrightarrow> z \<le> x \<Longrightarrow> sup y z \<le> x"
begin

lemma finite_set_has_LUB: "finite X \<Longrightarrow> X \<noteq> {} \<Longrightarrow> \<exists>b. b \<in> LUB X"
proof(induct X rule:finite_induct)
  case empty then show ?case by auto
next
  case IH: (insert x X)
  show ?case
  proof(cases "X = {}")
    case True
    with LUB_singleton have "LUB (insert x X) = [x]\<^sub>\<sim>" by auto
    then show ?thesis by auto
  next
    case False
    with IH have "LUB X \<noteq> {}" by auto
    then obtain b where bLUB: "b \<in> LUB X" by auto
    have "sup x b \<in> LUB (insert x X)"
    proof
      from bLUB show "sup x b \<in> UB (insert x X)" by(auto intro: sup_trans2)
    next
      fix c assume cUB: "c \<in> UB (insert x X)"
      with bLUB show "sup x b \<le> c" by(auto intro: sup_least)
    qed
    then show ?thesis by auto
  qed
qed

end

class quasi_order_bot = quasi_order + bot +
  assumes bot_least: "bot \<le> a"

text {* A class where the @{const Sup} operator is defined, which is valid only if it is definable. *}

class quasi_order_Sup = quasi_order + Sup +
  assumes Sup_in_LUB: "LUB X \<noteq> {} \<Longrightarrow> Sup X \<in> LUB X"
begin

lemma Sup_upper: assumes LUB: "LUB X \<noteq> {}" and xX: "x \<in> X" shows "x \<le> Sup X"
  using Sup_in_LUB[OF LUB] xX by auto

lemma Sup_least: assumes LUB: "LUB X \<noteq> {}" shows "(\<And>x. x \<in> X \<Longrightarrow> x \<le> y) \<Longrightarrow> Sup X \<le> y"
  using Sup_in_LUB[OF LUB] by auto

lemma Sup_equiv_LUB: assumes bLUB: "b \<in> LUB X" shows "Sup X \<sim> b"
  using Sup_in_LUB bLUB by auto

lemma Sup_singleton: "Sup {x} \<sim> x" by(rule Sup_equiv_LUB, auto)

lemma Sup_mono: assumes X: "LUB X \<noteq> {}" and Y: "LUB Y \<noteq> {}" and XY: "X \<subseteq> Y" shows "Sup X \<le> Sup Y"
  by (auto intro: LUB_mono[OF XY] Sup_in_LUB[OF X] Sup_in_LUB[OF Y])

end

text {* Upward complete quasi-semilattices -- where @{const Sup} is always defined. *}

class quasi_semilattice_Sup = quasi_order_Sup +
  assumes LUB: "LUB X \<noteq> {}"
begin

lemmas Sup_in_LUB = Sup_in_LUB[OF LUB]
lemmas Sup_upper = Sup_upper[OF LUB]
lemmas Sup_least = Sup_least[OF LUB]
lemmas Sup_mono = Sup_mono[OF LUB]

sublocale quasi_semilattice_sup "\<lambda>x y. Sup {x,y}"
  by (unfold_locales, auto intro: Sup_upper Sup_least)

sublocale quasi_order_bot "Sup {}"
proof(unfold_locales)
  fix x
  have "Sup {} \<le> Sup {x}" by (rule Sup_mono, auto)
  also have "... \<sim> x" by (rule Sup_singleton)
  finally show "Sup {} \<le> x".
qed

end

subsection {* @{term inv_imagep} *}

locale quasi_order_inv_imagep = base: quasi_order
begin

sublocale quasi_order "inv_imagep less_eq f" "inv_imagep less f"
  by (unfold_locales, unfold in_inv_imagep;
   fact base.order_refl base.order_trans base.less_trans base.less_le_trans base.le_less_trans)

end

locale wf_order_inv_imagep = base: wf_order
begin

interpretation quasi_order_inv_imagep..

sublocale wf_order "inv_imagep less_eq f" "inv_imagep less f"
  apply (unfold_locales)
  apply (unfold in_inv_imagep)
  using wf_inv_image[OF base.wf, unfolded wf_def] by auto

end

end